import sys, argparse

from pd_utils import parse_key, KeyList
from predictor import Predictor

def main():
    VERSION = '0.0.1a'

    parser = argparse.ArgumentParser(description='Predicts ASP logic program grounding sizes')
    parser.add_argument('files', nargs='+', default=[], help='gringo source files')
    parser.add_argument('--key', action='append', help='specifies the predicate keys to use (example format: p/3[1,3])')
    parser.add_argument('--predict_rules', type=str, help='a file containing rules to predict')
    # parser.add_argument('--histogram', action='store_true', help='')
    parser.add_argument('--version', action='store_true', help='displays the version of Predictor')
    parser.add_argument('--json', action='store_true', help='converts the output to JSON')

    args = parser.parse_args()
    if args.version:
        print('Predictor version %s' % VERSION)
        sys.exit(0)
    
    key_list = KeyList()
    if args.key is not None:
        for key_str in args.key:
            key_list.add_key(parse_key(key_str))
    predictor = Predictor(key_list)
    for file in args.files:
        predictor.load_program(file)
    if args.predict_rules is not None and args.predict_rules != '':
        raw_rules = predictor.read_file(args.predict_rules)
        prediction = predictor.predict_all(raw_rules)
    else:
        prediction = predictor.predict_loaded_program()
    if args.json:
        print('{\n\t"prediction": %s\n}' % prediction)
    else:
        print('Selectivity: %s' % predictor.analyzer.data_holder.selectivity)
        print('Program prediction is: %s' % prediction)

main()