import clingo_ast_util
# from clingo_ast_util import is_predicate_in_positive_body

def get_recursive_rules(predicates, module_rules):
    recursive_rules = []
    for rule in module_rules:
        if any(clingo_ast_util.is_predicate_in_positive_body(rule, p) for p in predicates):
            recursive_rules.append(rule)
    return recursive_rules

def get_exit_rules(predicates, module_rules, recursive_rules=None):
    if recursive_rules is None:
        recursive_rules = get_recursive_rules(predicates, module_rules)
    # Exit rules are those rules in the module that aren't recursive
    return [r for r in module_rules if r not in recursive_rules]

def is_int(s):
    try:
        int(s)
        return True
    except ValueError:
        return False

def parse_key(key_str):
    key_split = key_str.split('[')
    if len(key_split) != 2:
        raise ValueError('can\'t parse key predicate and key arguments from string (%s)' % key_str)
    # Predicate Symbol Parsing
    pred_str = key_split[0]
    pred_split = pred_str.split('/')
    if len(pred_split) != 2:
        raise ValueError('can\'t parse predicate name and arity from key string (%s, using predicate string %s)' % (key_str, pred_str))
    pred_sym_str = pred_split[0]
    arity = int(pred_split[1])
    pred_sym = clingo_ast_util.PredicateSymbol(pred_sym_str, arity)
    # Arguments Parsing
    args_str = key_split[1]
    if args_str[-1] != ']':
        raise ValueError('can\'t parse key arguments, missing closing bracket (%s)' % key_str)
    args_str = args_str[0:-1] # Remove closing bracket
    arg_indexes = [int(arg) for arg in args_str.split(',')]
    return Key(pred_sym, arg_indexes)

class Key(object):

    def __init__(self, pred, arg_index_list):
        self.pred = pred
        self.arg_index_list = arg_index_list # 0 indexed
    
    def is_index_keyed(self, index):
        return index in self.arg_index_list
    
    def __str__(self):
        return '%s%s' % (self.pred, self.arg_index_list)

class KeyList(object):

    def __init__(self, key_list=list()):
        self.keys = {}
        for key in key_list:
            self.add_key(key)

    def add_key(self, key):
        if key.pred in self.keys:
            raise Exception('Cannot add duplicate key: %s' % key)
        self.keys[key.pred] = key

    def get_key_for_predicate(self, pred):
        return self.keys.get(pred, None)
