from analyzer import Analyzer
# from clingo_ast_util import PredicateSymbol
import clingo_ast_util
from pd_utils import *
from networkx.algorithms.components.strongly_connected import strongly_connected_components

import sys
import clingo

class PredictorError(BaseException):
    '''Generic predictor exception'''
    pass

class ControlNotCreatedError(PredictorError):
    '''Raised when clingo control is accessed before its creation'''
    pass

debug = False

class Predictor():

    def __init__(self, key_list=KeyList()):
        self.settings = {
            'histogram': True,
            'keys': key_list
            # 'keys': KeyList([Key(clingo_ast_util.PredicateSymbol('p', 2), [0]), Key(clingo_ast_util.PredicateSymbol('q', 3), [0,1]), Key(clingo_ast_util.PredicateSymbol('r', 3), [0,1])])
            # 'keys': KeyList([Key(clingo_ast_util.PredicateSymbol('manAssignsScore', 3), [0,1]), Key(clingo_ast_util.PredicateSymbol('womanAssignsScore', 3), [0,1])])
            # 'keys': KeyList([Key(clingo_ast_util.PredicateSymbol('grid', 3), [0,1]), Key(clingo_ast_util.PredicateSymbol('number', 3), [0,1])])
            # 'keys': KeyList()
        }
        self.load_clingo()
        self.loaded_program = []
        self.analyzer = None

    def create_control(self):
        control = clingo.Control(['--warn=none'])
        control.use_enumeration_assumption = False
        # if pr_settings.args.instance:
        #     for instance_path in pr_settings.args.instance:
        #         control.load(instance_path)
        return control

    def load_clingo(self):
        self.control = self.create_control()
    
    def load_program(self, prog_file_path):
        '''Loads a program into the instance's knowledge base. If called multiple times, the new program will be added to the instance's knowledge base.'''
        raw_program = self.read_file(prog_file_path)
        with self.control.builder() as b:
            clingo.parse_program(raw_program, self._handle_parse)
    
    def _handle_parse(self, stm):
        self.loaded_program.append(stm)
    
    def predict_all(self, rules_str):
        rules = []
        clingo.parse_program(rules_str, lambda stm: rules.append(stm))
        return sum(self.predict_all_rules(rules))
    
    def predict_rule(self, rule):
        '''Predicts the grounding instantiation size of a rule based on the loaded program'''
        if self.analyzer is None:
            self.analyzer = Analyzer(self.control, self.settings)
            self.analyzer.analyze_program(self.loaded_program)
        dprint('IDB: %s' % self.analyzer.data_holder.idb_predicates())
        dprint('EDB: %s' % self.analyzer.data_holder.edb_predicates())
        dprint('EDB Tuples: %s' % self.analyzer.data_holder.edb)
        # dprint(self.analyzer.data_holder.edb)
        # dprint(self.analyzer.data_holder.edb_test)
        dprint('DG:')
        print_graph(self.analyzer.data_holder.dg)
        dprint('SCC: %s' % list(strongly_connected_components(self.analyzer.data_holder.dg)))
        dprint('CG:')
        print_graph(self.analyzer.data_holder.cg)
        dprint('ACS: %s' % self.analyzer.data_holder.acs)
        dprint('Selectivity Statistics: %s' % self.analyzer.data_holder.selectivity)
        dprint('Predicting %s' % rule)
        prediction = self.analyzer.predict_rule(rule)
        dprint('Prediction: %s' % prediction)
        return prediction
    
    def predict_all_rules(self, rule_list):
        if self.analyzer is None:
            self.analyzer = Analyzer(self.control, self.settings)
            self.analyzer.analyze_program(self.loaded_program)
        dprint('IDB: %s' % self.analyzer.data_holder.idb_predicates())
        dprint('EDB Tuples: %s' % self.analyzer.data_holder.edb.keys())
        dprint('EDB: %s' % self.analyzer.data_holder.edb_predicates())
        dprint('DG:')
        print_graph(self.analyzer.data_holder.dg)
        dprint('SCC: %s' % list(strongly_connected_components(self.analyzer.data_holder.dg)))
        dprint('CG:')
        print_graph(self.analyzer.data_holder.cg)
        dprint('ACS: %s' % self.analyzer.data_holder.acs)
        dprint('Selectivity Statistics: %s' % self.analyzer.data_holder.selectivity)
        dprint('Predicting %s' % rule_list)
        prediction = self.analyzer.predict_all_rules(rule_list)
        dprint('Predictions: %s' % prediction)
        return prediction
    
    def predict_loaded_program(self):
        if self.analyzer is None:
            self.analyzer = Analyzer(self.control, self.settings)
        self.analyzer.analyze_program(self.loaded_program)
        predictions = self.analyzer.predict_all_rules(self.loaded_program, update_statistics=False)
        dprint('DG:')
        print_graph(self.analyzer.data_holder.dg)
        dprint('CG:')
        print_graph(self.analyzer.data_holder.cg)
        dprint('Predictions: %s' % predictions)
        return sum(predictions)

    def read_file(self, file_path):
        with open(file_path) as f:
            content = f.read()
            return content

if __name__ == '__main__':
    raise PredictorError('Predictor must be imported as a module')

def print_graph(graph):
    for edge in graph.edges():
        graph_attr = graph[edge[0]][edge[1]]
        if 'label' in graph_attr:
            label = graph_attr['label']
            dprint('%s -[%s]-> %s' % (edge[0], label, edge[1]))
        else:
            dprint('%s ---> %s' % (edge[0], edge[1]))

def dprint(msg):
    if debug:
        print(msg)
